import { NavigationInstruction, RedirectToRoute, RouteConfig, Router, RouterConfiguration } from "aurelia-router";
import { IPitayaViewSettings, IScrollMarkerHandlerConfig, IViewStates }                     from "../interfaces.d";
import { navigate, PitayaApp, setStatusBarColor }                                           from "./app";
import { PitayaAppConfiguration }                                                           from "./app-configuration";
import { PitayaSubViewCustomElement }                                                       from "./components/pitaya-sub-view";
import { PitayaViewCustomElement }                                                          from "./components/pitaya-view";
import { log }                                                                              from "./logger";
import { generateGUID }                                                                     from "./utils";

const _CLASS_ANIMATION_READY: string  = "pitaya-view--animation-ready";
const _VIEW_ELEMENT_NAME: string      = "pitaya-view";
const _SUB_VIEW_ELEMENT_NAME: string  = "pitaya-sub-view";
const _TABS_VIEW_ELEMENT_NAME: string = "pitaya-tabs-view";
const _PITAYA_PAGES_CLASS: string     = "au-leave-active";

const _VIEW_STATUS_PRESET: IViewStates = {
	active   : false,
	created  : false,
	bound    : false,
	attached : false,
	animated : false,
	detached : false
};

const _MDC_COLOR_VARIABLE_MAP = {
	"primary"      : "--mdc-theme-primary",
	"secondary"    : "--mdc-theme-secondary",
	"on-primary"   : "--mdc-theme-on-primary",
	"on-secondary" : "--mdc-theme-on-secondary",
	"background"   : "--mdc-theme-background",
	"on-surface"   : "--mdc-theme-on-surface",
	"surface"      : "--mdc-theme-surface",
};

let _currentViewInstance: PitayaView    = null;
let _currentViewSubInstance: PitayaView = null;

export abstract class PitayaView
{
	public static readonly TABS_TRANSITION_LEFT: string  = "pitaya-tabs-transition__left";
	public static readonly TABS_TRANSITION_RIGHT: string = "pitaya-tabs-transition__right";

	static get instance()
	{
		return _currentViewInstance;
	}

	static get subInstance()
	{
		return _currentViewSubInstance;
	}

	private _routerConfig: RouterConfiguration = null;
	private _viewModel: PitayaViewCustomElement;
	private _animationTaskQueue: Function[]    = [];
	private _settings: IPitayaViewSettings     = {
		drawerItems         : [],
		drawerActiveItem    : "",
		drawerItemsDisabled : [],
		isPublic            : true,
		isSubView           : false,
		tabs                : [],
		hierarchy           : 0 as number,
		theme               : PitayaAppConfiguration.View.defaultTheme
	};

	private _states: IViewStates = Object.assign( {}, _VIEW_STATUS_PRESET );

	get states(): IViewStates
	{
		return this._states;
	}

	private _router: Router = null;

	get router(): Router
	{
		return this._router;
	}

	private _rootElement: HTMLElement;

	get rootElement(): HTMLElement
	{
		return this._rootElement;
	}

	private _innerElement: HTMLElement;

	get innerElement(): HTMLElement
	{
		return this._innerElement;
	}

	/**
	 * @aurelia hook
	 */
	public canActivate( params: any, routeConfig: RouteConfig, navigationInstruction: NavigationInstruction ): boolean | RedirectToRoute | Promise<boolean | RedirectToRoute>
	{
		const { View, Development } = PitayaAppConfiguration;

		if( this._settings.isSubView
			|| this._settings.isPublic
			|| ( Development.enabled && Development.disablePrivateRedirects )
			|| !View.privateRedirectView )
		{
			return true;
		}

		return new RedirectToRoute( View.privateRedirectView );
	}

	/**
	 * @aurelia hook
	 */
	public activate( params, routeConfig, navigationInstruction )
	{
		this._settings.hierarchy = routeConfig.nav || 0;

		this._states.active = true;
	}

	/**
	 * @aurelia hook
	 */
	public created( parentContext, context )
	{
		this._states.created = true;

		const model = context.controllers.find( ( controller ) =>
		{
			const { elementName } = controller.behavior;

			return elementName === _VIEW_ELEMENT_NAME
				|| elementName === _SUB_VIEW_ELEMENT_NAME
				|| elementName === _TABS_VIEW_ELEMENT_NAME;
		} );

		this._viewModel          = model.viewModel;
		this._settings.isSubView = this._viewModel instanceof PitayaSubViewCustomElement ? true : false;

		if( this._settings.isSubView )
		{
			_currentViewSubInstance = this;

			return;
		}

		_currentViewSubInstance = null;
		_currentViewInstance    = this;
	}

	/**
	 * @aurelia hook
	 */
	public bind( bindingContext, overrideContext )
	{
		this._states.bound = true;
	}

	/**
	 * @aurelia hook
	 */
	public attached()
	{
		this._states.attached = true;
		this._rootElement     = this._viewModel.rootElement as HTMLElement;
		this._innerElement    = this._viewModel.innerElement;

		this._settings.drawerActiveItem = this.constructor.name;

		this._setAnimationListener();
	}

	/**
	 * @aurelia hook
	 */
	public detached()
	{
		this._states.attached = false;
		this._states.detached = true;
	}

	public getNewId(): string
	{
		return generateGUID();
	}

	public async navigateSubRoute( routeName: string ): Promise<any>
	{
		return this._router.navigateToRoute( routeName );
	}

	public async navigate( route: string, options?: any )
	{
		return navigate( route, options );
	}

	public addAnimationTask( task: Function )
	{
		this._animationTaskQueue.push( task );
	}

	public onAnimationReady(){}

	public arrowBackTopbarAction()
	{
		PitayaApp.instance.navigateBack();
	}

	public setTheme( color: string = this._settings.theme )
	{
		setStatusBarColor( _getColorFromMdcVariable( color ) );
	}

	private _configureTabsRouter(): void
	{
		const routes = [];
		let index    = 0;

		for( const tab of this._settings.tabs )
		{
			const { route }   = tab;
			const routeConfig = index === 0 ? [ "", route ] : [ route ];

			routes.push( {
				route    : routeConfig,
				name     : route,
				moduleId : `./${route}`,
			} );

			index++;
		}

		( this as any ).configureRouter = function( config: RouterConfiguration, router: Router )
		{
			config.title = this.title;

			config.map( routes );
			config.mapUnknownRoutes( routes[ 0 ].moduleId );

			this._router       = router;
			this._routerConfig = config;
		};
	}

	private _willBeAnimated(): boolean
	{
		const { rootElement } = this;
		// @ts-ignore
		const siblings             = [ ...rootElement.parentElement.children ];

		return siblings.some( ( sibling ) => sibling !== rootElement && sibling.classList.contains( _PITAYA_PAGES_CLASS ) );
	}

	private async _setAnimationListener()
	{
		const { rootElement } = this;

		if( !this._willBeAnimated() )
		{
			this._onAnimationEnd();

			return;
		}

		rootElement.addEventListener( "animationend", this._onAnimationEnd.bind( this ), { once : true } );
	}

	private async _onAnimationEnd()
	{
		const queue = this._animationTaskQueue;

		while( queue.length )
		{
			await queue.pop()();
		}

		this.rootElement.classList.add( _CLASS_ANIMATION_READY );

		this._states.animated = true;

		!this._settings.isSubView && this.setTheme();

		this.onAnimationReady();
	}
}

export const scrollToTop = function( scrollDuration: number ): void
{
	const scrollStep     = -window.scrollY / ( scrollDuration / 15 );
	const scrollInterval = setInterval( function()
	{
		if( window.scrollY !== 0 )
		{
			window.scrollBy( 0, scrollStep );
		}
		else
		{
			clearInterval( scrollInterval );
		}
	}, 15 );
};

export const setScrollMarkerHandler = function( { element, className, minHeight, maxHeight }: IScrollMarkerHandlerConfig ): () => void
{
	const body               = document.body;
	const documentElement    = document.documentElement;
	const viewHeight: number = Math.max( documentElement.clientHeight, window.innerHeight || 0 );

	minHeight = minHeight === undefined ? 0 : minHeight;
	maxHeight = maxHeight === undefined ? viewHeight : maxHeight;

	const setScrollMarkerClass = function()
	{
		const scrollTop   = Math.max( body.scrollTop, documentElement.scrollTop );
		const addOrRemove = scrollTop >= minHeight && scrollTop <= maxHeight ? "add" : "remove";

		element.classList[ addOrRemove ]( className );
	};

	setScrollMarkerClass();

	window.addEventListener( "scroll", setScrollMarkerClass );

	return () =>
	{
		window.removeEventListener( "scroll", setScrollMarkerClass );
	};
};

const _getColorFromMdcVariable = function( color: string ): string
{
	if( color.startsWith( "#" ) || color.startsWith( "rgb" ) )
	{
		return color;
	}

	if( !( color in _MDC_COLOR_VARIABLE_MAP ) )
	{
		log.error( "view.js _getColorFromMdcVariable()", `No MDC variable found that matches '${color}'` );

		return "";
	}

	return getComputedStyle( PitayaApp.instance.appView.superViewElement ).getPropertyValue( _MDC_COLOR_VARIABLE_MAP[ color ] );
};